/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2004-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Needed for snprintf */
#define _XOPEN_SOURCE	500

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/x509_vfy.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/bio.h>
#include <openssl/asn1.h>
#include <openssl/evp.h>
#include <openssl/objects.h>

#include <openssl/rsa.h>
#include <openssl/des.h>

#include "_verify_log.h"
#include "_verify_x509.h"
#include "verify_x509.h"

#include "verify_x509_datatypes.h"


/* OpenSSL uses a default depth of 9.
 * To cope with Subordinate CAs we have to extend the verification depth to be
 * able to hold the certificate chain (could contain a lot of delegations) and
 * all the CA certificate, which might not be added to the certificate chain
 * itself but would still be lingering in the X509 CA directory lookup functions
 */
#define VERIFICATION_CHAIN_DEPTH_EXTENSION 9


#define SERIALSTR_LENGTH    ((size_t)255L)

#define EXT_TEXT_LEN	    80

/*****************************************************************************/
/* Internal variables                                                        */
/*****************************************************************************/

/**
 * library number to be set by ERR_get_next_error_library(). Can be initialized
 * using verify_init_library() and read out by verify_get_library().
 */
static int initialized=0;
static int library=-1;


/*****************************************************************************/
/* Internal prototypes                                                       */
/*****************************************************************************/

/**
 * converts proxy_type_t to human readable string
 */
static const char *grid_certificate_type_str(proxy_type_t cert_type);

/**
 * Callback function for PEM_read_bio_PrivateKey() and friends. See
 * PEM_read_bio_PrivateKey(3ssl)
 */
static int grid_X509_empty_password_callback(char *buf, int buf_size, int verify, void *cb_tmp);

/**
 * Callback function for OpenSSL to put the errors
 * Parameters: ok, X509_STORE_CTX
 * Returns: 1 on success, 0 on error
 */
static int grid_X509_verify_callback(int ok, X509_STORE_CTX *ctx);

/**
 * Checks the certificate chain on CA based (RFC5280) and RFC3820 Proxy based
 * Path Length Constraints.
 * Returns X509_V_OK on successful verification or X509_V_ERR_CERT_REJECTED on
 * any error.
 */
static int grid_verifyPathLenConstraints(STACK_OF(X509) * chain);

/**
 * Checks to see if one of the unsupported critical extensions is a valid proxy
 * type.
 * Returns X509_V_OK on success or X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION on
 * error
 */
static int grid_X509_knownCriticalExts(X509 *cert);

/**
 * Check if we can safely discards errors about the issuer
 * Returns 1 when issuer is fine, 0 when no.
 */
static int grid_X509_check_issued_wrapper(X509_STORE_CTX *ctx,X509 *x,X509 *issuer);

/**
 * Prints the expected types of certificates in human readable form.
 * Parameters:
 *   int expected_proxy_type          : bitfield of types of certificates that we expected.
 *   proxy_type_t got_type            : the (offending) certificate
 * Returns:
 *   char *message, MUST be free'd
 */
static char *grid_generate_proxy_expectation_error_message(proxy_type_t got_type, int expected_proxy_type);

/**
 * Converts the ASN1_INTEGER serial into a string
 * Returns a pointer to a static buffer or NULL on error
 */
static unsigned char *grid_get_serialStr(ASN1_INTEGER *serial);

/**
 * Tries to verify the proxies in the certstack
 * Return: X509_V_OK on success, or suitable X509_V_ERR_* code
 */
static int grid_verifyProxy( STACK_OF(X509) *certstack );


/*****************************************************************************/
/* Function definitions                                                      */
/*****************************************************************************/

/**
 * Converts an internal library 'reason' to an ERR_get_error() compatible code
 * and pushes the corresponding error on the stack.
 * Returns the error code
 * */
unsigned long verify_errval(int func, verify_x509_error_t reason,
			    const char *file, int line)	{
    /* Make sure the library is initialized */
    if (!initialized)
	verify_init_library();

    ERR_put_error(library, func, (int)reason, file, line); 
    return ERR_PACK(library,func, (int)reason);
}

/**
 * Converts an internal library 'reason' to an ERR_get_error() compatible code
 * and pushes the corresponding error on the stack.
 * Returns the reason code
 * */
verify_x509_error_t verify_reasonval(int func, verify_x509_error_t reason,
				     const char *file, int line)	{
    /* Make sure the library is initialized */
    if (!initialized)
	verify_init_library();

    ERR_put_error(library, func, (int)reason, file, line); 
    return reason;
}

/**
 * Initializes the library, by creating a new library ID and loading the error
 * and function strings.
 * Returns library number
 */
int verify_init_library(void)   {
    static ERR_STRING_DATA errstr[13],funcstr[5];

    /* Although in an ideal world we should use ERR_get_next_error_library(),
     * since VOMS is not using it, better set it by hand to a hopefully unused
     * value. */
/*    library=ERR_get_next_error_library();*/
    library=ERR_LIB_USER+64;

    /* Function code ==0 and reason ==0 sets the library name */
    errstr[0].error=ERR_PACK(library,0,0);
    errstr[0].string="Proxy Verification library";

    /* Function code ==0 and reason !=0 sets the reasons */
    errstr[1].error=ERR_PACK(library,0,VER_R_NO_CACERT);
    errstr[1].string="No CA certificate directory specified";
    errstr[2].error=ERR_PACK(library,0,VER_R_CERTSTACK_EMPTY);
    errstr[2].string="No certificate chain presented";
    errstr[3].error=ERR_PACK(library,0,VER_R_PARAMETER_EMPTY);
    errstr[3].string="Mandatory parameter is empty";
    errstr[4].error=ERR_PACK(library,0,VER_R_LIMITED_DISABLED);
    errstr[4].string="Limited proxies are disallowed by configuration";
    errstr[5].error=ERR_PACK(library,0,VER_R_NOPRIVATEKEY_DISABLED);
    errstr[5].string="Absence of private key disallowed by configuration";
    errstr[6].error=ERR_PACK(library,0,VER_R_X509_VERIFY_CERT_FAILURE);
    errstr[6].string="Certificate verification failed";

    errstr[7].error=ERR_PACK(library,0,VER_R_X509_PARAMS_CONTAINER_FAILURE);
    errstr[7].string="Parameter object is unset or empty";
    errstr[8].error=ERR_PACK(library,0,VER_R_X509_PARAMS_ALREADY_SET);
    errstr[8].string="Parameter is set multiple times";
    errstr[9].error=ERR_PACK(library,0,VER_R_X509_PARAMS_DATA_EMPTY);
    errstr[9].string="Parameter is set but value is empty";
    errstr[10].error=ERR_PACK(library,0,VER_R_X509_PARAMS_ACCESS_FAILURE);
    errstr[10].string="Parameter value cannot be accessed (I/O error)";
    errstr[11].error=ERR_PACK(library,0,VER_R_X509_PARAMS_UNSUPPORTED_DATATYPE);
    errstr[11].string="Unknown parameter type specified";

    errstr[12].error=0;
    errstr[12].string=NULL;

    /* Function code !=0 and reason ==0 sets the reasons */
    funcstr[0].error=ERR_PACK(library,VER_F_VERIFY_X509_VERIFY,0);
    funcstr[0].string="verify_x509_verify()";
    funcstr[1].error=ERR_PACK(library,VER_F_PROCESS_INTERNAL,0);
    funcstr[1].string="process_internal_verify_data()";
    funcstr[2].error=ERR_PACK(library,VER_F_GRID_VERIFYCERT,0);
    funcstr[2].string="grid_verifyCert()";
    funcstr[3].error=ERR_PACK(library,VER_F_SET_PARAM,0);
    funcstr[3].string="verify_X509_setParameter()";
    funcstr[4].error=0;
    funcstr[4].string=NULL;

    /* Store the error strings */
    ERR_load_strings(library,errstr);
    /* Store the function strings */
    ERR_load_strings(library,funcstr);

    initialized=1;

    return library;
}

/**
 * Reads the private key from filename and stores the result in pkey
 * Returns value of ERR_peek_error() on error or 0 on success.
 */
unsigned long verify_x509_readPrivateKeyFromFile(char *filename, EVP_PKEY **pkey)
{
    BIO                 *certbio = NULL;
    verify_log( L_DEBUG, "--- Reading the Private Key From File ---");

    if ( (certbio = BIO_new(BIO_s_file())) == NULL ) return ERR_peek_error();

    verify_log( L_INFO, "Reading file %s", filename );
    if ( BIO_read_filename(certbio, filename) <= 0 ) return ERR_peek_error();

    verify_log( L_DEBUG, "Reading Private key" );
    *pkey = PEM_read_bio_PrivateKey( certbio, NULL, grid_X509_empty_password_callback, NULL );

    if ( *pkey == NULL ) verify_log( L_WARN, "No private key found." );

    BIO_free(certbio);
    return 0;
}


/**
 * Reads the private key from PEM string and stores the result in pkey
 * Returns value of ERR_get_error() on error or 0 on success.
 */
unsigned long verify_x509_readPrivateKeyFromPEM (char * pem, EVP_PKEY **pkey)
{
    BIO                 *certbio = NULL;
    verify_log( L_DEBUG, "--- Reading the Private Key From PEM ---");

    /* if ( (certbio = BIO_new(BIO_s_mem())) == NULL ) return ERR_get_error(); */

    verify_log( L_INFO, "Reading PEM string", pem);
    if ( (certbio = BIO_new_mem_buf (pem, -1)) == NULL ) return ERR_peek_error();

    verify_log( L_DEBUG, "Reading Private key" );
    *pkey = PEM_read_bio_PrivateKey( certbio, NULL, grid_X509_empty_password_callback, NULL );

    if ( *pkey == NULL ) verify_log( L_WARN, "No private key found." );

    BIO_free(certbio);
    return 0;
}


/**
 * Reads the certificate chain from filename and stores the result in certstack
 * Returns value of ERR_get_error() on error or 0 on success.
 */
unsigned long verify_x509_readPublicCertChain(char *filename, STACK_OF(X509) **certstack)
{
    const char          *oper = "Reading proxy";

    STACK_OF(X509_INFO) *sk      = NULL;
    BIO                 *certbio = NULL;
    X509_INFO           *xi;
    unsigned long        err;

    verify_log( L_DEBUG, "--- Welcome to the verify_x509_readPublicCertChain function ---");

    *certstack = sk_X509_new_null();
    if (*certstack == NULL) return ERR_peek_error();

    if ( (certbio = BIO_new(BIO_s_file())) == NULL ) return ERR_peek_error();

    verify_log( L_INFO, "Reading file %s", filename );
    if ( BIO_read_filename(certbio, filename) <= 0 ) return ERR_peek_error();

    verify_log( L_DEBUG, "Reading X509_INFO records" );
    if ( !(sk=PEM_X509_INFO_read_bio(certbio, NULL, NULL, NULL)) )
    {
        err = ERR_peek_error();
        verify_error( oper, "No X509 records found" );
        BIO_free(certbio);
        sk_X509_INFO_free(sk);
        sk_X509_free(*certstack);
        return err;
    }

    while (sk_X509_INFO_num(sk))
    {
        xi=sk_X509_INFO_shift(sk);
        if (xi->x509 != NULL)
        {
            sk_X509_push(*certstack, xi->x509);
            xi->x509=NULL;
        }
        X509_INFO_free(xi);
    }

    if (!sk_X509_num(*certstack))
    {
        err = ERR_peek_error();
        verify_error( oper, "No certificates found" );
        BIO_free(certbio);
        sk_X509_INFO_free(sk);
        sk_X509_free(*certstack);
        return err;
    }

    BIO_free(certbio);
    sk_X509_INFO_free(sk);

    return 0;
}


#if 0
/******************************************************************************
Function:   verify_PROXYCERTINFO_get_path_length
Description:
            Get the proxy path length value
******************************************************************************/
long verify_PROXYCERTINFO_get_path_length(PROXYCERTINFO *cert_info) {
    if(cert_info && cert_info->path_length) {
        return ASN1_INTEGER_get(cert_info->path_length);
    } else {
        return -1;
    }
}
#endif


/**
 * Verifies a privatekey with the first cert in the certstack
 * returns 0 on success, or appropriate value from ERR_peek_error().
 */
unsigned long verify_verifyPrivateKey( STACK_OF(X509) *certstack, EVP_PKEY *pkey )
{
    X509 *cert = NULL;

    verify_log( L_DEBUG, "--- Welcome to the grid_verifyPrivateKey function ---");

    if ( pkey == NULL )
    {
        verify_log( L_WARN, "No private key available." );
        return 0;
    }

    /* Check for X509 certificate and point to it with 'cert' */
    if ((cert = sk_X509_value(certstack, 0)) != NULL)
    {
        verify_log( L_DEBUG, "X509_check_private_key" );
        if ( X509_check_private_key( cert, pkey ) != 1 )
        {
            return ERR_peek_error();
        }
    }

    return 0;
}


/**
 * Validates a certificate with the CRL list and the signing CA
 * Return 0 on success or result of ERR_peek_error() on error (to be fixed)
 */
unsigned long verify_verifyCert( char * CA_DIR, STACK_OF(X509) *certstack )
{
    const char     *oper = "Verifying certificate chain";

    X509_STORE     *store      = NULL;
    X509_LOOKUP    *lookup     = NULL;
    X509_STORE_CTX *verify_ctx = NULL;
    X509           *cert       = NULL;
    char *          cert_DN    = NULL;
    char *          issuer_DN  = NULL;
    int             i = 0;
    int             depth = 0;

    verify_log( L_DEBUG, "--- Welcome to the verify_verifyCert function ---");

    /* Initials must be good */
    if ( CA_DIR == NULL )
    {
        verify_error( oper, "No CA certificate directory specified." );
        return VERIFY_errval(VER_F_GRID_VERIFYCERT, VER_R_NO_CACERT);
    }
    if ( certstack == NULL )
    {
        verify_error( oper, "Certificate stack is empty." );
        return VERIFY_errval(VER_F_GRID_VERIFYCERT, VER_R_CERTSTACK_EMPTY);
    }

    verify_log( L_INFO, "Using CA Directory: %s", CA_DIR);

    verify_log( L_DEBUG, "X509_STORE_new");
    if (!(store = X509_STORE_new()))
    {
       verify_error( oper, "Could not create a X509 STORE." );
       return ERR_peek_error();
    }

    verify_log( L_DEBUG, "X509_STORE_set_verify_cb_func");
    X509_STORE_set_verify_cb_func (store, grid_X509_verify_callback);
#if 0
/* NOTE: the following two calls are defined in openssl ./crypto/x509/x509_d2.c
 * (at least for openssl-1.0.0f)
 * Also note that the X509_STORE_set_default_paths adds X509_LOOKUP_file (among
 * others) which is effectively the file pointed at by the env variable
 * SSL_CERT_FILE */
    /* Executing the lookups to the CA and CRL files */
    verify_log( L_DEBUG, "X509_STORE_load_locations");
    if (X509_STORE_load_locations (store, NULL, CA_DIR) != 1)
    {
        verify_error( oper, "Could not load the CA directory.");
        return ERR_get_error();
    }

    verify_log( L_DEBUG, "X509_STORE_set_default_paths");
    if (X509_STORE_set_default_paths(store) != 1)
    {
        verify_error( oper, "Could not load the system wide CA certificates.");
        return ERR_get_error();
    }
#endif

    /* Add CA_DIR, note: the next lines will effectively call dir_ctrl() from
     * ./crypto/x509/by_dir.c with type X509_FILETYPE_PEM */
    verify_log( L_DEBUG, "X509_STORE_add_lookup");
    if (!(lookup = X509_STORE_add_lookup (store, X509_LOOKUP_hash_dir())))
    {
        verify_error( oper, "Could not create X509_LOOKUP object.");
        return ERR_peek_error();
    }

    verify_log( L_DEBUG, "X509_LOOKUP_add_dir");
    i=X509_LOOKUP_add_dir (lookup, CA_DIR, X509_FILETYPE_PEM);
    if (!i)
    {
        verify_error( oper, "Could not add CA_DIR.");
        return ERR_peek_error();
    }

    verify_log( L_DEBUG, "X509_STORE_set_flags");

    store->check_issued = grid_X509_check_issued_wrapper;
#if OPENSSL_VERSION_NUMBER < 0x00908000L
    X509_STORE_set_flags( store, X509_V_FLAG_CRL_CHECK |
                                 X509_V_FLAG_CRL_CHECK_ALL );
#else
    X509_STORE_set_flags( store, X509_V_FLAG_CRL_CHECK |
                                 X509_V_FLAG_CRL_CHECK_ALL |
                                 X509_V_FLAG_ALLOW_PROXY_CERTS );
#endif

    verify_log( L_DEBUG, "X509_STORE_CTX_new");
    /* Creating a verification context and init it */
    if (!(verify_ctx = X509_STORE_CTX_new()))
    {
        verify_error( oper, "Could not create a X509 STORE CTX (context).");
        return ERR_peek_error();
    }

    depth = sk_X509_num (certstack);
    verify_log( L_DEBUG, "Depth of certstore %d", depth);

    for (i=depth-1; i >= 0; --i)
    {
        if ((cert = sk_X509_value(certstack, i)))
        {
            cert_DN   = X509_NAME_oneline(X509_get_subject_name(cert),NULL,0);
            issuer_DN = X509_NAME_oneline(X509_get_issuer_name(cert),NULL,0);

            verify_log( L_DEBUG, "DN[%d]:        %s", i, cert_DN );
            verify_log( L_DEBUG, "Issuer DN[%d]: %s", i, issuer_DN);

	    /* The X509_NAME_oneline() result without providing a buffer must be
	     * free'd */
            free(cert_DN);
            free(issuer_DN);

            if (verify_x509IsCA(cert))
            {
                verify_log( L_DEBUG, "This certificate is a CA certificate" );
                verify_log( L_DEBUG, "continue search for user certificate..." );
            }
            else
            {
                break;
            }
        }
    }

    cert = sk_X509_value( certstack, 0 );
    cert_DN   = X509_NAME_oneline( X509_get_subject_name( cert ) , NULL, 0 );
    issuer_DN = X509_NAME_oneline( X509_get_issuer_name( cert )  , NULL, 0 );

    verify_log( L_INFO, "Certificate to verify:" );
    verify_log( L_INFO, "  DN:        %s", cert_DN );
    verify_log( L_INFO, "  Issuer DN: %s", issuer_DN );

    /* The X509_NAME_oneline() result without providing a buffer must be free'd
     * */
    free(cert_DN);
    free(issuer_DN);


    verify_log( L_DEBUG, "X509_STORE_CTX_init" );
    if ( X509_STORE_CTX_init( verify_ctx, store, cert, certstack) != 1 )
    {
        verify_error( oper, "Could not initialize verification context.");
        return ERR_peek_error();
    }

    X509_STORE_CTX_set_purpose( verify_ctx, X509_PURPOSE_SSL_CLIENT );
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
    cert->ex_flags |= EXFLAG_PROXY;
#endif


    verify_log( L_DEBUG, "The certificate chain has a depth of %d. "
	    "For verification the depth is extended to fit the chain and "
	    "(subordinate) CAs to %d",
	    depth, depth + VERIFICATION_CHAIN_DEPTH_EXTENSION);

    /* Alter verification depth to fit the certificate chain, sub-CAs and root
     * CA */
    X509_STORE_CTX_set_depth (verify_ctx, depth + VERIFICATION_CHAIN_DEPTH_EXTENSION);


    verify_log( L_DEBUG, "X509_verify");

    if ( X509_verify_cert(verify_ctx) != 1)
    {
        verify_error( "Verifying certificate chain", "%s",
               X509_verify_cert_error_string(verify_ctx->error));
	/* Note: X509_verify_cert() already pushes a error on the error stack
	 * using X509err() (wrapper around ERR_PUT_error() defined in
	 * openssl/err.h), and we have just printed the X509_V_* error code, so
	 * we can suffice with returning that the verification failed */
	return VERIFY_errval(VER_F_GRID_VERIFYCERT, VER_R_X509_VERIFY_CERT_FAILURE);
    }
    else
    {
        verify_log( L_INFO, "The verification of the certicate has succeeded.");
    }

    if ( verify_ctx ) X509_STORE_CTX_free( verify_ctx );
    if ( store )      X509_STORE_free( store );

    return 0;
}




/******************************************************************************
 * Internal functions
 *****************************************************************************/

/**
 * converts proxy_type_t to human readable string
 */
static const char *grid_certificate_type_str(proxy_type_t cert_type) {
    return cert_type == CA ? "CA" :
           cert_type == EEC ? "EEC" :
           cert_type == GT2_PROXY ? "GT2/old-style Proxy" :
           cert_type == GT3_PROXY ? "GT3/pre-RFC Proxy" :
           cert_type == RFC_PROXY ? "RFC3820 Proxy" :
           cert_type == GT2_LIMITED_PROXY ? "GT2/old-style Limited Proxy" :
           cert_type == GT3_LIMITED_PROXY ? "GT3/pre-RFC Limited Proxy" :
           cert_type == RFC_LIMITED_PROXY ? "RFC3820 Limited Proxy" :
           "Unknown";
}

/**
 * Callback function for PEM_read_bio_PrivateKey() and friends. See
 * PEM_read_bio_PrivateKey(3ssl)
 */
static int grid_X509_empty_password_callback(char *buf, int buf_size, int verify, void *cb_tmp)
{
         if ( buf_size > 0 ) *buf = '\0';
              return 0;
}


/**
 * Callback function for OpenSSL to put the errors
 * Parameters: ok, X509_STORE_CTX
 * Returns: 1 on success, 0 on error
 */
static int grid_X509_verify_callback(int ok, X509_STORE_CTX *ctx)
{
    int errnum   = X509_STORE_CTX_get_error(ctx);
    int errdepth = X509_STORE_CTX_get_error_depth(ctx);
    STACK_OF(X509) *certstack;

    /* When not ok... */
    if (ok != 1)
    {
        verify_log( L_DEBUG,
	    "verification callback indicated \'ok = %d\' and error number = %d:",
	    ok, errnum);

        if (errnum == X509_V_ERR_INVALID_CA)	{
	    verify_log( L_INFO, "Invalid CA.\n");
	    ok=1;
	}
        if (errnum == X509_V_ERR_UNABLE_TO_GET_CRL)
	{
	    verify_log( L_INFO, "Unable to get CRL.\n");
	    ok=1;
	}
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
        /* I don't want to do this, really, but I have yet to figure out
           how to get openssl 0.9.8 to accept old-style proxy certificates...
        */
        if (errnum == X509_V_ERR_INVALID_PURPOSE)   {
	    verify_log( L_INFO, "Invalid purpose.\n");
	    ok=1;
	}
#endif
        if (errnum == X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION)
        {
	    verify_log( L_DEBUG, "Unhandled critical extension, running grid_X509_knownCriticalExts() to check if we know it.\n");
            errnum = grid_X509_knownCriticalExts(X509_STORE_CTX_get_current_cert(ctx));
            ctx->error = errnum;
            if (errnum == X509_V_OK) ok=1;
        }

        /* Path length exceeded for the CA (should never happen in OpenSSL - famous last words) */
        if (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED)
        {
            verify_log( L_DEBUG, "Shallow Error X509_V_ERR_PATH_LENGTH_EXCEEDED: Running alternative RFC5280 and RFC3820 compliance tests.");
            if ((ctx->error = grid_verifyPathLenConstraints(X509_STORE_CTX_get_chain(ctx))) != X509_V_OK)
                goto failure;
	    /* It's ok */
	    ok=1;
        }

        /* Path length exceeded for the Proxy! -> Override and continue */
        /* This is NOT about X509_V_ERR_PATH_LENGTH_EXCEEDED */
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
        if (ctx->error == X509_V_ERR_PROXY_PATH_LENGTH_EXCEEDED)
        {
            verify_log( L_DEBUG, "Shallow Error X509_V_ERR_PROXY_PATH_LENGTH_EXCEEDED: Running alternative RFC5280 and RFC3820 compliance tests.");
            if ((ctx->error = grid_verifyPathLenConstraints(X509_STORE_CTX_get_chain(ctx))) != X509_V_OK)
                goto failure;
	    /* It's ok */
	    ok=1;
        }
#endif
    }

    /*
     * We've now got the last certificate - the identity being used for
     * this connection. At this point we check the whole chain for valid
     * CAs or, failing that, GSI-proxy validity using grid_verifyProxy
     */
    if ((errdepth == 0) && (ok == 1)) {
        certstack = (STACK_OF(X509) *) X509_STORE_CTX_get_chain( ctx );

        if ((ctx->error = grid_verifyProxy(certstack)) != X509_V_OK) {
            goto failure;
        } else {
            /* I'm in the OK state - Let's see if the chain is really well formed */
            if ((ctx->error = grid_verifyPathLenConstraints(X509_STORE_CTX_get_chain(ctx))) != X509_V_OK)
                goto failure;
        }
    }

    /* Ready and probably ok */
    return ok;

failure:
    /* Indicate failure */
    ok = 0;
    verify_log( L_INFO, "grid_verify_callback: error code: %d, message: \"%s\"", ctx->error, X509_verify_cert_error_string (ctx->error));

    return ok;
}

/**
 * Checks the certificate chain on CA based (RFC5280) and RFC3820 Proxy based
 * Path Length Constraints.
 * Returns X509_V_OK on successful verification or X509_V_ERR_CERT_REJECTED on
 * any error.
 */
static int grid_verifyPathLenConstraints(STACK_OF(X509) * chain)
{
    const char *oper = "grid_verifyPathLenConstraints";
    X509 *cert = NULL;
    int i, depth;
    proxy_type_t curr_cert_type = NONE;
    /* Note: cannot define expe_cert_type as proxy_type_t since it's a
     * combination of them */
    int expe_cert_type = CA|EEC|RFC_PROXY|GT2_PROXY;
    int found_EEC = 0;
    char *cert_subjectdn = NULL;
    char *error_msg = NULL;

    long ca_path_len_countdown    = -1;
    long proxy_path_len_countdown = -1;

    /* No chain, no game */
    if (!chain) {
        verify_error( oper, "No certificate chain detected.");
        goto failure;
    }

    /* Go through the list, from the CA(s) down through the EEC to the final delegation */
    depth = sk_X509_num (chain);
    for (i=depth-1; i >= 0; --i) {
        if ((cert = sk_X509_value(chain, i))) {
            /* Init to None, indicating not to have identified it yet */
            curr_cert_type = NONE;

            /* Extract Subject DN - Needs free */
            if (!(cert_subjectdn = X509_NAME_oneline (X509_get_subject_name (cert), NULL, 0))) {
                verify_error (oper, "Couldn't get the subject DN from the certificate at depth %d", depth);
                goto failure;
            }

            /* verify_log (L_DEBUG, "\tCert here is: %s", cert_subjectdn); */
            curr_cert_type = verify_type_of_proxy(cert);
            if (curr_cert_type == NONE) {
                verify_error (oper, "Couldn't classify certificate at depth %d with subject DN \"%s\"",
                             depth, cert_subjectdn);
                goto failure;
            }

            /* Mark that we've found an EEC - When we see it again, it's a failure */
            if (curr_cert_type == EEC && found_EEC == 0) {
                found_EEC = 1;
            } else if (curr_cert_type == EEC && found_EEC == 1) {
                verify_error (oper, "Found another EEC classified certificate in the same chain at depth %d with subject DN \"%s\"",
                             depth, cert_subjectdn);
                goto failure;
            }


#if 0
                /* NOTE: This is for quick debugging only */
                error_msg = grid_generate_proxy_expectation_error_message(curr_cert_type, expe_cert_type);
                printf("%s: Build chain checker: %s. Cert at depth %d of %d with Subject DN: %s",
                            oper,
                            error_msg,
                            i,
                            depth,
                            cert_subjectdn);
                free(error_msg); error_msg = NULL;
#endif

            /* Expectation management */
            if (!((expe_cert_type & (int)curr_cert_type) == (int)curr_cert_type)) {
                /* Failed to comply with the expectations! */
#define USE_STRICT_PATH_VALIDATION
#ifdef USE_STRICT_PATH_VALIDATION
                error_msg = grid_generate_proxy_expectation_error_message(curr_cert_type, expe_cert_type);
                verify_error(oper, "Certificate chain not build in the right order. %s. Cert at depth %d of %d with Subject DN: %s",
                            error_msg,
                            i,
                            depth,
                            cert_subjectdn);
                free(error_msg); error_msg = NULL;
                goto failure;
#else
                error_msg = grid_generate_proxy_expectation_error_message(curr_cert_type, expe_cert_type);
                verify_log(L_INFO, "%s: Certificate chain not build in the right order. %s. Cert at depth %d of %d with Subject DN: %s",
                            oper,
                            error_msg,
                            i,
                            depth,
                            cert_subjectdn);
                free(error_msg); error_msg = NULL;
                goto continue_after_warning;
#endif
            }
#ifndef USE_STRICT_PATH_VALIDATION
continue_after_warning:
#endif

            if (curr_cert_type == CA) {
                /* Expected next certificate type is: CA or EEC certificate */
                expe_cert_type = CA|EEC;
                verify_log (L_DEBUG, "Current cert is a CA: %s", cert_subjectdn);

                /* Exceeded CA Path Length ? */
                if (ca_path_len_countdown == 0) {
                    verify_error(oper, "CA Path Length Constraint exceeded on depth %d for certificate \"%s\". No CA certifcates were expected at this stage.", i, cert_subjectdn);
                    goto failure;
                }

                /* Store pathlen, override when small, otherwise keep the smallest */
                if (cert->ex_pathlen != -1) {
                    /* Update when ca_path_len_countdown is the initial value
                     * or when the PathLenConstraint is smaller then the
                     * remembered ca_path_len_countdown */
                    if ((ca_path_len_countdown == -1) || (cert->ex_pathlen < ca_path_len_countdown)) {
                        ca_path_len_countdown = cert->ex_pathlen;
                    } else {
                        /* If a path length was already issuesd, lower ca_path_len_countdown */
                        if (ca_path_len_countdown != -1)
                            ca_path_len_countdown--;
                    }
                } else {
                    /* If a path length was already issuesd, lower ca_path_len_countdown */
                    if (ca_path_len_countdown != -1)
                        ca_path_len_countdown--;
                }

            } else if (curr_cert_type == EEC) {
                /* Expected next certificate type is: GT2_PROXY, GT3_PROXY, RFC_PROXY or a Limited proxy of these flavors certificate */
                expe_cert_type = GT2_PROXY|GT3_PROXY|RFC_PROXY|GT2_LIMITED_PROXY|GT3_LIMITED_PROXY|RFC_LIMITED_PROXY;
                verify_log (L_DEBUG, "Current cert is a EEC: %s", cert_subjectdn);

            } else if (curr_cert_type == GT2_PROXY) {
                /* Expected next certificate type is: GT2_PROXY certificate */
                expe_cert_type = GT2_PROXY|GT2_LIMITED_PROXY;
                verify_log (L_DEBUG, "Current cert is a GT2 Proxy: %s", cert_subjectdn);

            } else if (curr_cert_type == GT2_LIMITED_PROXY) {
                /* Expected next certificate type is: GT2_LIMITED_PROXY certificate */
                expe_cert_type = GT2_LIMITED_PROXY;
                verify_log (L_DEBUG, "Current cert is a GT2 Limited Proxy: %s", cert_subjectdn);

            } else if (curr_cert_type == GT3_PROXY) {
                /* Expected next certificate type is: GT3_PROXY certificate */
                expe_cert_type = GT3_PROXY|GT3_LIMITED_PROXY;
                verify_log (L_DEBUG, "Current cert is a GT3 Proxy: %s", cert_subjectdn);
            } else if (curr_cert_type == GT3_LIMITED_PROXY) {
                /* Expected next certificate type is: GT3_LIMITED_PROXY certificate */
                expe_cert_type = GT3_LIMITED_PROXY;
                verify_log (L_DEBUG, "Current cert is a GT3 Limited Proxy: %s", cert_subjectdn);

            } else if (curr_cert_type == RFC_PROXY) {
                /* Expected next certificate type is: RFC_PROXY certificate */
                expe_cert_type = RFC_PROXY|RFC_LIMITED_PROXY;
                verify_log (L_DEBUG, "Current cert is a RFC Proxy: %s", cert_subjectdn);

                /* Exceeded CA Path Length ? */
                if (proxy_path_len_countdown == 0) {
                    verify_error(oper, "Proxy Path Length Constraint exceeded on depth %d of %d for certificate \"%s\". No Proxy certifcates were expected at this stage.", i, depth, cert_subjectdn);
                    goto failure;
                }

/* OpenSSL pre-0.9.8 has no proxy support and hence no ex_pcpathlen in X509 */
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
                /* Store pathlen, override when small, otherwise keep the smallest */
                if (cert->ex_pcpathlen != -1) {
                    /* Update when proxy_path_len_countdown is the initial value
                     * or when the PathLenConstraint is smaller then the
                     * remembered proxy_path_len_countdown */

                    if ((proxy_path_len_countdown == -1) || (cert->ex_pcpathlen < proxy_path_len_countdown)) {
                        proxy_path_len_countdown = cert->ex_pcpathlen;
                        verify_log (L_DEBUG, "Cert here is: %s -> Setting proxy path len constraint to: %d", cert_subjectdn, cert->ex_pcpathlen);
                    } else {
                        /* If a path length was already issuesd, lower ca_path_len_countdown */
                        if (proxy_path_len_countdown != -1)
                            proxy_path_len_countdown--;

                        verify_log (L_DEBUG, "Cert here is: %s -> Countdown is at %d", cert_subjectdn, proxy_path_len_countdown);
                    }
                } else {
#endif
                    /* If a path length was already issued, lower ca_path_len_countdown */
                    if (proxy_path_len_countdown != -1) {
                        proxy_path_len_countdown--;
                        verify_log (L_DEBUG, "Cert here is: %s -> Countdown is at %d", cert_subjectdn, proxy_path_len_countdown);
                    }

#if OPENSSL_VERSION_NUMBER >= 0x00908000L
                }
#endif
            } else if (curr_cert_type == RFC_LIMITED_PROXY) {
                /* Expected next certificate type is: RFC_LIMITED_PROXY certificate */
                expe_cert_type = RFC_LIMITED_PROXY;
                verify_log (L_DEBUG, "Current cert is a RFC Limited Proxy: %s", cert_subjectdn);

                /* Exceeded CA Path Length ? */
                if (proxy_path_len_countdown == 0) {
                    verify_error(oper, "Proxy Path Length Constraint exceeded on depth %d of %d for certificate \"%s\". No Proxy certifcates were expected at this stage.", i, depth, cert_subjectdn);
                    goto failure;
                }

/* OpenSSL pre-0.9.8 has no proxy support and hence no ex_pcpathlen in X509 */
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
                /* Store pathlen, override when small, otherwise keep the smallest */
                if (cert->ex_pcpathlen != -1) {
                    /* Update when proxy_path_len_countdown is the initial value
                     * or when the PathLenConstraint is smaller then the
                     * remembered proxy_path_len_countdown */

                    if ((proxy_path_len_countdown == -1) || (cert->ex_pcpathlen < proxy_path_len_countdown)) {
                        proxy_path_len_countdown = cert->ex_pcpathlen;
                        verify_log (L_DEBUG, "Cert here is: %s -> Setting proxy path len constraint to: %d", cert_subjectdn, cert->ex_pcpathlen);
                    } else {
                        /* If a path length was already issuesd, lower ca_path_len_countdown */
                        if (proxy_path_len_countdown != -1)
                            proxy_path_len_countdown--;

                        verify_log (L_DEBUG, "Cert here is: %s -> Countdown is at %d", cert_subjectdn, proxy_path_len_countdown);
                    }
                } else {
#endif
                    /* If a path length was already issued, lower ca_path_len_countdown */
                    if (proxy_path_len_countdown != -1) {
                        proxy_path_len_countdown--;
                        verify_log (L_DEBUG, "Cert here is: %s -> Countdown is at %d", cert_subjectdn, proxy_path_len_countdown);
                    }

#if OPENSSL_VERSION_NUMBER >= 0x00908000L
                }
#endif
            }

            /* Free memory during each cycle */
            if (cert_subjectdn) {
                free(cert_subjectdn);
                cert_subjectdn = NULL;
            }
        }
    }

/*success:*/
    /* Return an OK (thumbs up) in the grid_X509_verify_callback() */
    if (cert_subjectdn) {
        free(cert_subjectdn);
        cert_subjectdn = NULL;
    }
    return X509_V_OK;

failure:
    if (cert_subjectdn) {
        free(cert_subjectdn);
        cert_subjectdn = NULL;
    }
    return X509_V_ERR_CERT_REJECTED;
}



/**
 * Checks to see if one of the unsupported critical extensions is a valid proxy
 * type.
 * Returns X509_V_OK on success or X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION on
 * error
 */
static int grid_X509_knownCriticalExts(X509 *cert)
{
   int  i;
   char s[EXT_TEXT_LEN];
   X509_EXTENSION *ex;

   for (i = 0; i < X509_get_ext_count(cert); ++i)
      {
        ex = X509_get_ext(cert, i);

        if (X509_EXTENSION_get_critical(ex) &&
                                 !X509_supported_extension(ex))
          {
            OBJ_obj2txt(s, EXT_TEXT_LEN, X509_EXTENSION_get_object(ex), 1);

            verify_log( L_DEBUG, "Critical extension found: %s", s );

            if (strcmp(s, OID_RFC_PROXY) == 0) return X509_V_OK;
            if (strcmp(s, OID_GLOBUS_PROXY_V3) == 0) return X509_V_OK;
            if (strcmp(s, OID_GLOBUS_PROXY_V2) == 0) return X509_V_OK;

            return X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION;
          }
      }

   return X509_V_OK;
}

/**
 * Check if we can safely discards errors about the issuer
 * Returns 1 when issuer is fine, 0 when no.
 */
static int grid_X509_check_issued_wrapper(X509_STORE_CTX *ctx,X509 *x,X509 *issuer)
/* We change the default callback to use our wrapper and discard errors
   due to GSI proxy chains (ie where users certs act as CAs) */
{
    int ret;
    ret = X509_check_issued(issuer, x);
    if (ret == X509_V_OK) return 1;

    /* Non self-signed certs without signing are ok if they passed
           the other checks inside X509_check_issued. Is this enough? */
    if ((ret == X509_V_ERR_KEYUSAGE_NO_CERTSIGN) &&
        (X509_NAME_cmp(X509_get_subject_name(issuer),
                       X509_get_subject_name(x)) != 0)) return 1;

#if OPENSSL_VERSION_NUMBER < 0x00908000L
    /* If we haven't asked for issuer errors don't set ctx */
    if (!(ctx->flags & X509_V_FLAG_CB_ISSUER_CHECK)) return 0;
#else
    if (!(ctx->param->flags & X509_V_FLAG_CB_ISSUER_CHECK)) return 0;
#endif

    ctx->error = ret;
    ctx->current_cert = x;
    ctx->current_issuer = issuer;
    /* verify_cb() returns 1 on ok. */
    return ctx->verify_cb(0, ctx);
}



/**
 * Prints the expected types of certificates in human readable form.
 * Parameters:
 *   proxy_type_t expected_proxy_type : bitfield of types of certificates that we expected.
 *   proxy_type_t got_type            : the (offending) certificate
 * Returns:
 *   char *message, MUST be free'd
 */
static char *grid_generate_proxy_expectation_error_message(proxy_type_t got_type, int expected_proxy_type) {
    char *msg = NULL;
    const char *got_sub_msg = NULL;
    char *expected_sub_msg = NULL;
    size_t len;
    int  expectations_cnt = 0;
    size_t expectations_len = 0;
    const char *expectation_delimiter = " or ";

    if (expected_proxy_type & CA) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(CA));
        } else {                     expectations_len += strlen(grid_certificate_type_str(CA)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }
    if (expected_proxy_type & EEC) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(EEC));
        } else {                     expectations_len += strlen(grid_certificate_type_str(EEC)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT2_PROXY) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(GT2_PROXY));
        } else {                     expectations_len += strlen(grid_certificate_type_str(GT2_PROXY)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT3_PROXY) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(GT3_PROXY));
        } else {                     expectations_len += strlen(grid_certificate_type_str(GT3_PROXY)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }
    if (expected_proxy_type & RFC_PROXY) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(RFC_PROXY));
        } else {                     expectations_len += strlen(grid_certificate_type_str(RFC_PROXY)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT2_LIMITED_PROXY) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(GT2_LIMITED_PROXY));
        } else {                     expectations_len += strlen(grid_certificate_type_str(GT2_LIMITED_PROXY)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT3_LIMITED_PROXY) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(GT3_LIMITED_PROXY));
        } else {                     expectations_len += strlen(grid_certificate_type_str(GT3_LIMITED_PROXY)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }
    if (expected_proxy_type & RFC_LIMITED_PROXY) {
        if (expectations_cnt == 0) { expectations_len += strlen(grid_certificate_type_str(RFC_LIMITED_PROXY));
        } else {                     expectations_len += strlen(grid_certificate_type_str(RFC_LIMITED_PROXY)) + strlen(expectation_delimiter); }
        expectations_cnt++;
    }

    /* Creating the buffer for the humanly readable exected list of proxy types */
    expected_sub_msg = malloc(expectations_len + 1);
    if (expected_sub_msg == NULL) {
        return NULL;
    }
    memset(expected_sub_msg, 0, expectations_len + 1);
    expectations_cnt = 0;

    if (expected_proxy_type & CA) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(CA));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(CA));
        }
        expectations_cnt++;
    }
    if (expected_proxy_type & EEC) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(EEC));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(EEC));
        }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT2_PROXY) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(GT2_PROXY));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(GT2_PROXY));
        }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT3_PROXY) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(GT3_PROXY));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(GT3_PROXY));
        }
        expectations_cnt++;
    }
    if (expected_proxy_type & RFC_PROXY) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(RFC_PROXY));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(RFC_PROXY));
        }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT2_LIMITED_PROXY) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(GT2_LIMITED_PROXY));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(GT2_LIMITED_PROXY));
        }
        expectations_cnt++;
    }
    if (expected_proxy_type & GT3_LIMITED_PROXY) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(GT3_LIMITED_PROXY));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(GT3_LIMITED_PROXY));
        }
        expectations_cnt++;
    }
    if (expected_proxy_type & RFC_LIMITED_PROXY) {
        if (expectations_cnt == 0) { strcpy(expected_sub_msg, grid_certificate_type_str(RFC_LIMITED_PROXY));
        } else {                     strcat(expected_sub_msg, expectation_delimiter);
                                     strcat(expected_sub_msg, grid_certificate_type_str(RFC_LIMITED_PROXY));
        }
        expectations_cnt++;
    }

    got_sub_msg=grid_certificate_type_str(got_type);
    /* length format=101 - 4 for 2x%s + 1 for \0 */
    len=98+strlen(got_sub_msg)+strlen(expected_sub_msg);
    if ( (msg=(char*)malloc(len)) == NULL)
	goto finalize;
    snprintf(msg, len, "Certificate chain not build in the right order. Got %s certificate, but expected a(n) %s certificate.",
                        got_sub_msg,expected_sub_msg);

finalize:
    free(expected_sub_msg);
    return msg;
}


/**
 * Converts the ASN1_INTEGER serial into a string
 * Returns a pointer to a static buffer or NULL on error
 */
static unsigned char *grid_get_serialStr(ASN1_INTEGER *serial) {
    const char *oper = "grid_get_serialStr";
    unsigned char *temp;
    unsigned char serialNumberDER[127];
    static unsigned char serialStr[SERIALSTR_LENGTH];
    int serialLen, j = 0;
    int ret;

    /* Get length and fill serialNumberDER */
    temp=serialNumberDER;
    serialLen = i2c_ASN1_INTEGER( serial, &temp );

    /* Clear serialStr buffer */
    memset(serialStr,'\0',sizeof( serialStr));

    temp = serialStr;
    for (j = 0; j < serialLen; j++)
    {
	ret=snprintf( (char*)temp, SERIALSTR_LENGTH, "%02X", serialNumberDER[j] );
	if (ret<0)	{
	    verify_error( oper, "Error converting serialNumberDER: %s",strerror(errno));
	    return NULL;
	}
	/* Note: ret>=0 by now */
	if ((size_t)ret>=SERIALSTR_LENGTH)	{
	    verify_error( oper, "Error converting serialNumberDER, buffer too small");
	    return NULL;
	}

	temp += 2;
    }
    return serialStr;
}

/**
 * Tries to verify the proxies in the certstack
 * Return: X509_V_OK on success, or suitable X509_V_ERR_* code
 */
static int grid_verifyProxy( STACK_OF(X509) *certstack )
{
    const char *oper = "Verifying proxy";

    int      i = 0;
    X509    *cert = NULL;
    time_t   now = time((time_t *)NULL);
    size_t   len = 0;             /* Lengths of issuer and cert DN */
    size_t   len2 = 0;            /* Lengths of issuer and cert DN */
    int      prevIsLimited = 0;   /* previous cert was proxy and limited */
    char    *cert_DN = NULL;      /* Pointer to current-certificate-in-certstack's DN */
    char    *issuer_DN = NULL;    /* Pointer to issuer-of-current-cert-in-certstack's DN */
    char    *proxy_part_DN = NULL;
    int      depth = sk_X509_num (certstack);
    int      amount_of_CAs = 0;
    int      is_old_style_proxy = 0;
    int      is_limited_proxy = 0;
    ASN1_INTEGER   *cert_Serial = NULL;
    ASN1_INTEGER   *issuer_Serial = NULL;
    unsigned char  *serialStr;
    EVP_PKEY *pkey = NULL;
    int	     key_strength;

    verify_log( L_DEBUG, "--- Welcome to the grid_verifyProxy function ---");

    /* And there was (current) time... */
    time(&now);

    /* How many CA certs are there in the certstack? */
    for (i = 0; i < depth; i++)
    {
        if (verify_x509IsCA(sk_X509_value(certstack, i)))
            amount_of_CAs++;
    }

    verify_log( L_DEBUG, "#CA's = %d , depth = %d", amount_of_CAs, depth );

    if ((amount_of_CAs + 2) > depth)
    {
	if ((depth - amount_of_CAs) > 0) /* i.e. depth==amount_of_CAs+1,
					    following from previous if combined
					    with this one */
        {
            verify_log( L_WARN, "No proxy certificate in certificate stack to check." );
            return X509_V_OK;
        }
        else
        {
            verify_error( oper, "No personal certificate (neither proxy or user certificate) found in the certficiate stack." );
            return X509_V_ERR_APPLICATION_VERIFICATION;
        }
    }


    /* Changed this value to start checking the proxy and such and
       to skip the CA and the user_cert
    */
    for (i = depth - (amount_of_CAs + 2); i >= 0; i--)
    {
        /* Check for X509 certificate and point to it with 'cert' */
        if ( (cert = sk_X509_value(certstack, i)) != NULL )
        {
            cert_DN   = X509_NAME_oneline( X509_get_subject_name( cert), NULL, 0);
            issuer_DN = X509_NAME_oneline( X509_get_issuer_name( cert ), NULL, 0);
            len       = strlen( cert_DN );
            len2      = strlen( issuer_DN );

            verify_log( L_INFO, "Proxy to verify:" );
            verify_log( L_INFO, "  DN:        %s", cert_DN );
            verify_log( L_INFO, "  Issuer DN: %s", issuer_DN );

            if (now < verify_asn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notBefore(cert))))
            {
                verify_error( oper, "Proxy certificate is not yet valid." );
                return X509_V_ERR_CERT_NOT_YET_VALID;
            }

            if (now > verify_asn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notAfter(cert))))
            {
                verify_error( oper, "Proxy certificate expired." );
/* error will be caught later by x509_verify
                return X509_V_ERR_CERT_HAS_EXPIRED;
*/
            }

            /* User not allowed to sign shortened DN */
            if (len2 > len)
            {
                verify_error( oper, "It is not allowed to sign a shorthened DN.");
                return X509_V_ERR_APPLICATION_VERIFICATION;
            }

            /* Proxy subject must begin with issuer. */
            if (strncmp(cert_DN, issuer_DN, len2) != 0)
            {
                verify_error( oper, "Proxy subject must begin with the issuer.");
                return X509_V_ERR_APPLICATION_VERIFICATION;
            }

            /* Set pointer to end of base DN in cert_DN */
            proxy_part_DN = &cert_DN[len2];

            /* First attempt at support for Old and New style GSI
               proxies: /CN=anything is ok for now */
            if (strncmp(proxy_part_DN, "/CN=", (size_t)4) != 0)
            {
                verify_error( oper, "Could not find a /CN= structure in the DN, thus it is not a proxy.");
                return X509_V_ERR_APPLICATION_VERIFICATION;
            }

            if (strncmp(proxy_part_DN, "/CN=proxy", (size_t)9) == 0)
            {
                verify_log( L_INFO, "Current certificate is an old style proxy.");
                is_old_style_proxy = 1;
                is_limited_proxy = 0;
            }
            else if (strncmp(proxy_part_DN, "/CN=limited proxy", (size_t)17) == 0)
            {
                verify_log( L_INFO, "Current certificate is an old limited style proxy.");
                is_old_style_proxy = 1;
                is_limited_proxy = 1;
            }
            else
            {
                verify_log( L_INFO, "Current certificate is a GSI/RFC3820 proxy.");
            }

            if ( is_old_style_proxy )
            {
                cert_Serial = X509_get_serialNumber( cert );
		if ( (serialStr=grid_get_serialStr(cert_Serial)) )
		    verify_log( L_DEBUG, "Serial number: %s", serialStr);

                issuer_Serial = X509_get_serialNumber( sk_X509_value(certstack, i+1));
		if ( (serialStr=grid_get_serialStr(issuer_Serial)) )
		    verify_log( L_DEBUG, "Issuer serial number: %s", serialStr);

                if (cert_Serial && issuer_Serial)
                {
                    if ( ASN1_INTEGER_cmp( cert_Serial, issuer_Serial ) )
                    {
                        verify_log( L_WARN, "Serial numbers do not match." );
                    }
                }
            }

            if ( is_limited_proxy )
            {
                prevIsLimited = 1;
                if (i > 0) verify_log( L_WARN, "Found limited proxy.");
            }
            else
            {
                if (prevIsLimited)
                {
                    verify_error( oper, "Proxy chain integrity error. Previous proxy in chain was limited, but this one is a regular proxy.");
                    return X509_V_ERR_APPLICATION_VERIFICATION;
                }
            }

            /* Now check the public key strength */
            pkey=X509_get_pubkey( cert );
            if (pkey == NULL)
            {
                verify_error( oper, "Cannot retrieve public key" );
            }
            else
            {
                if (pkey->type == EVP_PKEY_RSA)
                {
                    key_strength = BN_num_bits(pkey->pkey.rsa->n);

                    verify_log( L_DEBUG, "Key strength = %d", key_strength );
                    if (key_strength < 1024)
			verify_log( L_WARN, "Key strength too low (%d).", key_strength );
                }
                else
                { 
                    verify_log( L_DEBUG, "No RSA public key found?!?" );
                }
            }

            if (cert_DN) free(cert_DN);
            if (issuer_DN) free(issuer_DN);
        }
    }

    return X509_V_OK;
}


